jail = {}

local keyboardmap = setmetatable({
	[" "] = "space"
}, {__index = function(self, key) return key end})
local mousemap = setmetatable({
	wu = "mouse_wheelup",
	wd = "mouse_wheeldown"
}, {__index = function(self, key) return "mouse_" .. key end})
local joystickmap = setmetatable({
}, {__index = function(self, key) return key end})

local keymap = {}
local keys = {}
local modules = {
	keyboard = true,
	mouse = true,
	joystick = true
}
local deadzone = 0.1
local hook_table
local hook_self = false
local keyfile
local debug

local function parsefile()
	for line in love.filesystem.lines(keyfile) do
		local key, action = line:match("^bind (%S+) (%S+)")
		if key and action then
			local pos, neg = key:match("^(%S+):(%S+)$")
			if pos and neg then
				if action:match("^%-") then
					keymap[pos] = action
					keymap[neg] = action:sub(2, -1)
				else
					keymap[pos] = action
					keymap[neg] = "-" .. action
				end
				if debug then
					print(string.format("keymap[%q] = %q", pos, keymap[pos]))
					print(string.format("keymap[%q] = %q", neg, keymap[neg]))
				end
			else
				keymap[key] = action
				if debug then print(string.format("keymap[%q] = %q", key, action)) end
			end
		end
	end
end

local keyboardfunctions = [=[
	local t, map, keymap, keys, pressed, released = ...
	function t%skeypressed(key)
		key = keymap[map[key]]
		if not key then return end
		pressed(key)
	end

	function t%skeyreleased(key)
		key = keymap[map[key]]
		if not key then return end
		released(key)
	end
]=]

local mousefunctions = [=[
	local t, map, keymap, keys, pressed, released = ...
	function t%smousepressed(x, y, key)
		key = keymap[map[key]]
		if not key then return end
		pressed(key)
	end

	function t%smousereleased(x, y, key)
		key = keymap[map[key]]
		if not key then return end
		released(key)
	end
]=]

local joystickfunctions = [=[
	local t, map, keymap, keys, pressed, released  = ...
	function t%sjoystickpressed(num, key)
		key = keymap[map[("j%%d_%%d"):format(num, key)]]
		if not key then return end
		pressed(key)
	end

	function t%sjoystickreleased(num, key)
		key = keymap[map[("j%%d_%%d"):format(num, key)]]
		if not key then return end
		released(key)
	end
]=]

local function pressed(key)
	local fkey = key:match("^%-(.+)$")
	if fkey then
		keys[fkey] = -1
	else
		keys[key] = 1
	end
	if not hook_table.pressed then return end
	if hook_self then
		hook_table:pressed(key)
	else
		hook_table.pressed(key)
	end
end

local function released(key)
	local fkey = key:match("^%-(.+)$")
	if fkey then
		keys[fkey] = 0
	else
		keys[key] = 0
	end
	if not hook_table.released then return end
	if hook_self then
		hook_table:released(key)
	else
		hook_table.released(key)
	end
end

function jail.init(settings)
	debug = settings.debug
	assert(settings.keymap, "No keymap set!")
	keyfile = settings.keymap
	assert(settings.keys, "No keys set!")
	for i, v in ipairs(settings.keys) do
		keys[v] = 0
	end
	if not love.filesystem.exists(keyfile) then
		assert(settings.default and love.filesystem.exists(settings.default), "No default keymap set or it doesn't exist!")
		love.filesystem.write(keyfile, love.filesystem.read(settings.default))
	end
	parsefile()
	if settings.hook ~= false then
		settings.hook_table = settings.hook_table or love
		hook_table = settings.hook_table
		hook_self = settings.hook_self
		local sep = hook_self and ":" or "."
		if settings.keyboard ~= false then
			loadstring(keyboardfunctions:format(sep, sep))(hook_table, keyboardmap, keymap, keys, pressed, released)
		else
			modules.keyboard = false
		end
		if settings.mouse ~= false then
			loadstring(mousefunctions:format(sep, sep))(hook_table, mousemap, keymap, keys, pressed, released)
		else
			modules.mouse = false
		end
		if settings.joystick ~= false then
			loadstring(joystickfunctions:format(sep, sep))(hook_table, joystickmap, keymap, keys, pressed, released)
		else
			modules.joystick = false
		end
	end
end

function jail.isDown(key)
	return keys[key] and keys[key] ~= 0
end

function jail.getDown(key)
	return keys[key]
end

function jail.update()
	for joystick = 0, love.joystick.getNumJoysticks()-1 do
		if not love.joystick.isOpen(joystick) then
			love.joystick.open(joystick)
		end
		for axis = 0, love.joystick.getNumAxes(joystick)-1 do
			local val = love.joystick.getAxis(joystick, axis)
			local negative = val < 0
			val = math.abs(val)-deadzone
			val = (val < 0 and 0 or val)/(1-deadzone)
			if negative then val = -val end
			local key = keymap[("j%d_a%d"):format(joystick, axis)]
			if key and keys[key] then
				local oldval = keys[key]
				if math.abs(keys[key]) < 2*deadzone or val ~= 0 then
					keys[key] = val
				end
				if oldval == 0 and keys[key] == 1 then
					pressed(key)
				elseif oldval == 1 and keys[key] == 0 then
					released(key)
				end
			end
		end
	end
end
